@@ -1,3 +1,4 @@
+#include "libfilezilla/buffer.hpp"
 #include "libfilezilla/encode.hpp"
 
 namespace fz {
@@ -67,6 +68,13 @@ std::string base64_encode(std::vector<uint8_t> const& in, base64_type type, bool
 	return ret;
 }
 
+std::string base64_encode(fz::buffer const& in, base64_type type, bool pad)
+{
+	std::string ret;
+	base64_encode_impl(ret, in.to_view(), type, pad);
+	return ret;
+}
+
 void base64_encode_append(std::string& result, std::string_view const& in, base64_type type, bool pad)
 {
 	base64_encode_impl(result, in, type, pad);
@@ -172,6 +180,11 @@ std::vector<uint8_t> base64_decode(std::wstring_view const& in)
 	return base64_decode_impl<std::vector<uint8_t>>(in);
 }
 
+std::vector<uint8_t> base64_decode(fz::buffer const& in)
+{
+	return base64_decode_impl<std::vector<uint8_t>>(in.to_view());
+}
+
 std::string base64_decode_s(std::string_view const& in)
 {
 	return base64_decode_impl<std::string>(in);
@@ -182,6 +195,11 @@ std::string base64_decode_s(std::wstring_view const& in)
 	return base64_decode_impl<std::string>(in);
 }
 
+std::string base64_decode_s(fz::buffer const& in)
+{
+	return base64_decode_impl<std::string>(in.to_view());
+}
+
 
 namespace {
 template<typename DataContainer>
@@ -275,6 +293,10 @@ std::string base32_encode(std::vector<uint8_t> const& in, base32_type type, bool
 	return base32_encode_impl(in, type, pad);
 }
 
+std::string base32_encode(fz::buffer const& in, base32_type type, bool pad)
+{
+	return base32_encode_impl(in.to_view(), type, pad);
+}
 
 namespace {
 template<typename Ret, typename View>
@@ -435,6 +457,11 @@ std::vector<uint8_t> base32_decode(std::wstring_view const& in, base32_type type
 	return base32_decode_impl<std::vector<uint8_t>>(in, type);
 }
 
+std::vector<uint8_t> base32_decode(fz::buffer const& in, base32_type type)
+{
+	return base32_decode_impl<std::vector<uint8_t>>(in.to_view(), type);
+}
+
 std::string base32_decode_s(std::string_view const& in, base32_type type)
 {
 	return base32_decode_impl<std::string>(in, type);
@@ -445,6 +472,11 @@ std::string base32_decode_s(std::wstring_view const& in, base32_type type)
 	return base32_decode_impl<std::string>(in, type);
 }
 
+std::string base32_decode_s(fz::buffer const& in, base32_type type)
+{
+	return base32_decode_impl<std::string>(in.to_view(), type);
+}
+
 
 std::string percent_encode(std::string_view const& s, bool keep_slashes)
 {
