@@ -170,7 +170,7 @@ inline std::wstring to_wstring(T const& in) {
 
 /// Converts from arithmetic type to std::wstring
 template<typename Arg>
-inline typename std::enable_if<std::is_arithmetic<std::decay_t<Arg>>::value, std::wstring>::type to_wstring(Arg && arg)
+inline typename std::enable_if<std::is_arithmetic_v<std::decay_t<Arg>>, std::wstring>::type to_wstring(Arg && arg)
 {
 	return std::to_wstring(std::forward<Arg>(arg));
 }
@@ -205,7 +205,7 @@ inline std::string to_string(T const& in) {
 
 /// Converts from arithmetic type to std::string
 template<typename Arg>
-inline typename std::enable_if<std::is_arithmetic<std::decay_t<Arg>>::value, std::string>::type to_string(Arg && arg)
+inline typename std::enable_if<std::is_arithmetic_v<std::decay_t<Arg>>, std::string>::type to_string(Arg && arg)
 {
 	return std::to_string(std::forward<Arg>(arg));
 }
@@ -236,13 +236,13 @@ std::string FZ_PUBLIC_SYMBOL to_utf8(std::wstring_view const& in);
 
 /// Calls either fz::to_string or fz::to_wstring depending on the passed template argument
 template<typename String, typename Arg>
-inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same<String, std::string>::value, decltype(to_string(std::forward<Arg>(arg)))>::type
+inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same_v<String, std::string>, decltype(to_string(std::forward<Arg>(arg)))>::type
 {
 	return to_string(std::forward<Arg>(arg));
 }
 
 template<typename String, typename Arg>
-inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same<String, std::wstring>::value, decltype(to_wstring(std::forward<Arg>(arg)))>::type
+inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same_v<String, std::wstring>, decltype(to_wstring(std::forward<Arg>(arg)))>::type
 {
 	return to_wstring(std::forward<Arg>(arg));
 }
@@ -285,15 +285,25 @@ template<> inline wchar_t const* choose_string(char const*, wchar_t const* w) {
 #define fzS(Char, s) fz::choose_string<Char>(s, L ## s)
 #endif
 
- /// Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
-std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string const& in, std::string_view const& find, std::string_view const& replacement);
-std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring const& in, std::wstring_view const& find, std::wstring_view const& replacement);
-std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string const& in, char find, char replacement);
-std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring const& in, wchar_t find, wchar_t replacement);
+ /** \brief Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
+  *
+  * \arg find If empty, no replacement takes place.
+  */
+std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string_view const& in, std::string_view const& find, std::string_view const& replacement);
+std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring_view const& in, std::wstring_view const& find, std::wstring_view const& replacement);
 
-/// Modifies \c in, replacing all occurrences of \c find with \c replacement
+/// Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
+std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string_view const& in, char find, char replacement);
+std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring_view const& in, wchar_t find, wchar_t replacement);
+
+/** \brief Modifies \c in, replacing all occurrences of \c find with \c replacement
+ *
+ * \arg find If empty, no replacement takes place.
+ */
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::string& in, std::string_view const& find, std::string_view const& replacement);
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::wstring& in, std::wstring_view const& find, std::wstring_view const& replacement);
+
+/// Modifies \c in, replacing all occurrences of \c find with \c replacement
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::string& in, char find, char replacement);
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::wstring& in, wchar_t find, wchar_t replacement);
 
@@ -333,7 +343,10 @@ inline auto FZ_PUBLIC_SYMBOL strtok_view(std::wstring_view const& tokens, wchar_
 template<typename T, typename String>
 T to_integral_impl(String const& s, T const errorval = T())
 {
-	if constexpr (std::is_enum_v<T>) {
+	if constexpr (std::is_same_v<T, bool>) {
+		return static_cast<T>(to_integral_impl<unsigned int>(s, static_cast<unsigned int>(errorval))) != 0;
+	}
+	else if constexpr (std::is_enum_v<T>) {
 		return static_cast<T>(to_integral_impl<std::underlying_type_t<T>>(s, static_cast<std::underlying_type_t<T>>(errorval)));
 	}
 	else {
@@ -527,6 +540,14 @@ bool ends_with(String const& s, String const& ending)
 	}
 }
 
+/**
+ * Normalizes various hyphens, dashes and minuses to just hyphen-minus.
+ *
+ * The narrow version assumes UTF-8 as encoding.
+ */
+std::string FZ_PUBLIC_SYMBOL normalize_hyphens(std::string_view const& in);
+std::wstring FZ_PUBLIC_SYMBOL normalize_hyphens(std::wstring_view const& in);
+
 }
 
 #endif
