@@ -9,7 +9,7 @@
 #include <limits>
 
 #ifdef FZ_WINDOWS
-#include "private/windows.hpp"
+#include "glue/windows.hpp"
 #endif
 
 /** \file
@@ -32,7 +32,7 @@ equivalent to a timestamp received via the protocol.
 
 The internal representation of \c datetime is in milliseconds since 1970-01-01 00:00:00.000 UTC and can handle a
 range of several million years. While datetime supports negative times (i.e. earlier than 1970-01-01 00:00:00.000 UTC),
-the underlying plaform may not support it.
+the underlying platform may not support it.
 
 \remark Some *nix systems base their time on TAI instead of UTC, though we pretend there
 is no difference, as the latter is the default on every modern distribution.
@@ -61,7 +61,7 @@ public:
 	};
 
 	/// A default-constructed timestamp is \ref empty()
-	datetime() = default;
+	datetime() noexcept = default;
 
 	datetime(zone z, int year, int month, int day, int hour = -1, int minute = -1, int second = -1, int millisecond = -1);
 
@@ -69,10 +69,10 @@ public:
 
 	/**
 	 * \brief Construct from string, looks for \c YYYYmmDD[[[[HH]MM]SS]sss]
-	 * \see bool set(std::string const& str, zone z)
+	 * \see bool set(std::string_view const& str, zone z)
 	 */
-	explicit datetime(std::string const& s, zone z);
-	explicit datetime(std::wstring const& s, zone z);
+	explicit datetime(std::string_view const& s, zone z);
+	explicit datetime(std::wstring_view const& s, zone z);
 
 #ifdef FZ_WINDOWS
 	/// Windows-only: Construct from FILETIME
@@ -110,6 +110,7 @@ public:
 	bool operator<(datetime const& op) const;
 	bool operator<=(datetime const& op) const;
 	bool operator>(datetime const& op) const { return op < *this; }
+	bool operator>=(datetime const& op) const { return op <= *this; }
 	/// \}
 
 	/**
@@ -131,7 +132,7 @@ public:
 
 	/** \name Adding/subtracting duration intervals
 	 *
-	 * Adding or subracting a \ref duration interval is accuracy-aware, e.g. adding a single second to a datetime with
+	 * Adding or subtracting a \ref duration interval is accuracy-aware, e.g. adding a single second to a datetime with
 	 * minute-accuracy does not change the timestamp.
 	 */
 	datetime& operator+=(duration const& op);
@@ -161,8 +162,8 @@ public:
 	 *
 	 * \return whether setting the timestamp succeeded. On failure object gets cleared
 	 */
-	bool set(std::string const& str, zone z);
-	bool set(std::wstring const& str, zone z);
+	bool set(std::string_view const& str, zone z);
+	bool set(std::wstring_view const& str, zone z);
 
 #ifdef FZ_WINDOWS
 	/// Windows-only: Set timestamp from FILETIME
@@ -245,8 +246,22 @@ public:
 	 * Sun, 06-Nov-94 08:49:37 GMT // obsolete RFC 850\n
 	 * Sun Nov 6 08:49:37 1994 // ANSI C asctime\n
 	 */
-	bool set_rfc822(std::string const& str);
-	bool set_rfc822(std::wstring const& str);
+	bool set_rfc822(std::string_view const& str);
+	bool set_rfc822(std::wstring_view const& str);
+
+	/**
+	 * Parses a date in the format specified in RFC 3339
+	 *
+	 * On errors, object gets clears
+	 *
+	 * \return whether setting the timestamp succeeded. On failure object gets cleared.
+	 *
+	 * \par Examples:
+	 * 1985-04-12T23:20:50.52Z
+	 * 1996-12-19T16:39:57-08:00
+	 */
+	bool set_rfc3339(std::string_view const& str);
+	bool set_rfc3339(std::wstring_view const& str);
 
 private:
 	int FZ_PRIVATE_SYMBOL compare_slow(datetime const& op) const;
@@ -267,11 +282,13 @@ private:
  * have the time unit as part of the function name.
  *
  * In contract to \ref datetime, \c duration does not track accuracy.
+ * 
+ * \note Arithmetic operations on duration do not check for integer over/underflow
  */
 class FZ_PUBLIC_SYMBOL duration final
 {
 public:
-	duration() = default;
+	duration() noexcept = default;
 
 	/** \name Getters
 	 * All getters return the total time of the duration, rounded down to the requested granularity.
@@ -301,6 +318,11 @@ public:
 	}
 	/// \}
 
+	duration& operator+=(duration const& op) {
+		ms_ += op.ms_;
+		return *this;
+	}
+
 	duration& operator-=(duration const& op) {
 		ms_ -= op.ms_;
 		return *this;
@@ -314,14 +336,24 @@ public:
 		return ms_ != 0;
 	}
 
+	duration& operator*=(int64_t op) {
+		ms_ *= op;
+		return *this;
+	}
+
+	duration absolute() const {
+		return (ms_ < 0) ? duration(-ms_) : *this;
+	}
+
 	bool operator<(duration const& op) const { return ms_ < op.ms_; }
 	bool operator<=(duration const& op) const { return ms_ <= op.ms_; }
 	bool operator>(duration const& op) const { return ms_ > op.ms_; }
 	bool operator>=(duration const& op) const { return ms_ >= op.ms_; }
 
 	friend duration FZ_PUBLIC_SYMBOL operator-(duration const& a, duration const& b);
+	friend duration FZ_PUBLIC_SYMBOL operator+(duration const& a, duration const& b);
 private:
-	explicit duration(int64_t ms) : ms_(ms) {}
+	explicit FZ_PRIVATE_SYMBOL duration(int64_t ms) : ms_(ms) {}
 
 	int64_t ms_{};
 };
@@ -331,6 +363,10 @@ inline duration operator-(duration const& a, duration const& b)
 	return duration(a) -= b;
 }
 
+inline duration operator+(duration const& a, duration const& b)
+{
+	return duration(a) += b;
+}
 
 /** \relates datetime
  * \relatesalso duration
@@ -396,7 +432,7 @@ public:
 	}
 
 private:
-	explicit monotonic_clock(clock_type::time_point const& t)
+	explicit FZ_PRIVATE_SYMBOL monotonic_clock(clock_type::time_point const& t)
 		: t_(t)
 	{}
 
