@@ -4,7 +4,9 @@
 #include "libfilezilla/file.hpp"
 
 #ifdef FZ_WINDOWS
+#include "windows/dll.hpp"
 #include "windows/security_descriptor_builder.hpp"
+#include <winternl.h>
 #else
 #include <errno.h>
 #include <sys/fcntl.h>
@@ -39,7 +41,7 @@ local_filesys::~local_filesys()
 
 namespace {
 #ifdef FZ_WINDOWS
-	bool IsNameSurrogateReparsePoint(std::wstring const& file)
+bool IsNameSurrogateReparsePoint(std::wstring const& file)
 {
 	WIN32_FIND_DATA data;
 	HANDLE hFind = FindFirstFile(file.c_str(), &data);
@@ -372,34 +374,29 @@ result local_filesys::begin_find_files(native_string path, bool dirs_only, bool
 	end_find_files();
 
 	if (path.empty()) {
-		return result{result::nodir};
+		return {result::invalid};
 	}
 
 	dirs_only_ = dirs_only;
 	query_symlink_targets_ = query_symlink_targets;
 
 #ifdef FZ_WINDOWS
-	if (is_separator(path.back())) {
-		m_find_path = path;
-		path += '*';
-	}
-	else {
-		m_find_path = path + fzT("\\");
-		path += fzT("\\*");
+	if (!is_separator(path.back())) {
+		path += '\\';
 	}
 
-	m_hFind = FindFirstFileEx(path.c_str(), FindExInfoStandard, &m_find_data, dirs_only ? FindExSearchLimitToDirectories : FindExSearchNameMatch, nullptr, 0);
-	if (m_hFind == INVALID_HANDLE_VALUE) {
-		has_next_ = false;
-		switch (GetLastError()) {
-			case ERROR_ACCESS_DENIED:
-				return result{result::noperm};
-			default:
-				return result{result::other};
+	dir_ = CreateFile(path.c_str(), FILE_LIST_DIRECTORY, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, nullptr, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, nullptr);
+	if (dir_ == INVALID_HANDLE_VALUE) {
+		auto const err = GetLastError();
+		switch (err) {
+		case ERROR_ACCESS_DENIED:
+			return result{result::noperm, err};
+		default:
+			return result{result::other, err};
 		}
 	}
 
-	has_next_ = true;
+	buffer_.resize(64 * 1024);
 	return result{result::ok};
 #else
 	if (path.size() > 1 && path.back() == '/') {
@@ -408,23 +405,40 @@ result local_filesys::begin_find_files(native_string path, bool dirs_only, bool
 
 	dir_ = opendir(path.c_str());
 	if (!dir_) {
-		switch (errno) {
+		int const err = errno;
+		switch (err) {
 			case EACCES:
 			case EPERM:
-				return result{result::noperm};
+				return {result::noperm, err};
 			case ENOTDIR:
 			case ENOENT:
-				return result{result::nodir};
+				return {result::nodir, err};
 			default:
-				return result{result::other};
+				return {result::other, err};
 		}
 	}
 
-	return result{result::ok};
+	return {result::ok};
 #endif
 }
 
-#if FZ_UNIX
+#if FZ_WINDOWS
+result local_filesys::begin_find_files(HANDLE dir, bool dirs_only, bool query_symlink_targets)
+{
+	end_find_files();
+
+	if (dir == INVALID_HANDLE_VALUE) {
+		return {result::nodir};
+	}
+
+	dirs_only_ = dirs_only;
+	query_symlink_targets_ = query_symlink_targets;
+
+	dir_ = dir;
+	buffer_.resize(64 * 1024);
+	return {result::ok};
+}
+#else
 result local_filesys::begin_find_files(int fd, bool dirs_only, bool query_symlink_targets)
 {
 	end_find_files();
@@ -438,15 +452,17 @@ result local_filesys::begin_find_files(int fd, bool dirs_only, bool query_symlin
 
 	dir_ = fdopendir(fd);
 	if (!dir_) {
-		switch (errno) {
+		close(fd);
+		int const err = errno;
+		switch (err) {
 			case EACCES:
 			case EPERM:
-				return result{result::noperm};
+				return {result::noperm, err};
 			case ENOTDIR:
 			case ENOENT:
-				return result{result::nodir};
+				return {result::nodir, err};
 			default:
-				return result{result::other};
+				return {result::other, err};
 		}
 	}
 
@@ -457,11 +473,12 @@ result local_filesys::begin_find_files(int fd, bool dirs_only, bool query_symlin
 void local_filesys::end_find_files()
 {
 #ifdef FZ_WINDOWS
-	has_next_ = false;
-	if (m_hFind != INVALID_HANDLE_VALUE) {
-		FindClose(m_hFind);
-		m_hFind = INVALID_HANDLE_VALUE;
+	if (dir_ != INVALID_HANDLE_VALUE) {
+		CloseHandle(dir_);
+		dir_ = INVALID_HANDLE_VALUE;
 	}
+	buffer_.clear();
+	cur_ = nullptr;
 #else
 	if (dir_) {
 		closedir(dir_);
@@ -470,30 +487,138 @@ void local_filesys::end_find_files()
 #endif
 }
 
+#ifdef FZ_WINDOWS
+namespace {
+extern "C" {
+// Sadly ddk/ntifs.h is unusuable. Manually declare the needed things.
+typedef enum
+{
+	lfzFileDirectoryInformation = 1
+} lfzFILE_INFORMATION_CLASS;
+
+typedef struct {
+	ULONG NextEntryOffset;
+	ULONG unused1;
+	LARGE_INTEGER CreationTime;
+	LARGE_INTEGER LastAccessTime;
+	LARGE_INTEGER LastWriteTime;
+	LARGE_INTEGER ChangeTime;
+	LARGE_INTEGER EndOfFile;
+	LARGE_INTEGER unused2;
+	ULONG FileAttributes;
+	ULONG FileNameLength;
+	WCHAR FileName[1];
+} lfzFILE_DIRECTORY_INFORMATION;
+
+typedef NTSTATUS(NTAPI* lfzNtQueryDirectoryFile)(HANDLE dir, HANDLE ev, void*, void*, IO_STATUS_BLOCK* status, void* buffer, ULONG size, lfzFILE_INFORMATION_CLASS c, BOOL single, void*, BOOL restart);
+typedef NTSTATUS(NTAPI* lfzNtOpenFile)(HANDLE* file, ACCESS_MASK DesiredAccess, OBJECT_ATTRIBUTES* attributes, IO_STATUS_BLOCK* status, ULONG ShareAccess, ULONG OpenOptions);
+}
+
+bool read_dir_buffer(HANDLE dir, void* buf, ULONG size, lfzFILE_INFORMATION_CLASS c)
+{
+	static dll ntdll(L"ntdll.dll", LOAD_LIBRARY_SEARCH_SYSTEM32);
+	static lfzNtQueryDirectoryFile f = ntdll.h_ ? reinterpret_cast<lfzNtQueryDirectoryFile>(GetProcAddress(ntdll.h_, "NtQueryDirectoryFile")) : nullptr;
+	if (!f) {
+		return false;
+	}
+	IO_STATUS_BLOCK s{};
+	NTSTATUS res = f(dir, nullptr, nullptr, nullptr, &s, buf, size, c, FALSE, nullptr, FALSE);
+	if (res != 0 || !s.Information) {
+		return false;
+	}
+	return true;
+}
+
+HANDLE OpenAt(HANDLE dir, std::wstring const& name, ACCESS_MASK DesiredAccess, ULONG ShareAccess, ULONG OpenOptions)
+{
+	static dll ntdll(L"ntdll.dll", LOAD_LIBRARY_SEARCH_SYSTEM32);
+	static lfzNtOpenFile const func = ntdll.h_ ? reinterpret_cast<lfzNtOpenFile>(GetProcAddress(ntdll.h_, "NtOpenFile")) : nullptr;
+	if (!func) {
+		return INVALID_HANDLE_VALUE;
+	}
+
+	UNICODE_STRING uname;
+	uname.Length = name.size() * 2;
+	uname.MaximumLength = name.size() * 2 + 2;
+	uname.Buffer = const_cast<wchar_t*>(name.c_str());
+
+	OBJECT_ATTRIBUTES attrs{};
+	attrs.Length = sizeof(OBJECT_ATTRIBUTES);
+	attrs.RootDirectory = dir;
+	attrs.ObjectName = &uname;
+
+	HANDLE file;
+	IO_STATUS_BLOCK s{};
+	NTSTATUS res = func(&file, DesiredAccess | SYNCHRONIZE, &attrs, &s, ShareAccess, OpenOptions | FILE_SYNCHRONOUS_IO_NONALERT);
+	if (res) {
+		return INVALID_HANDLE_VALUE;
+	}
+	return file;
+}
+
+bool IsNameSurrogateReparsePoint(HANDLE dir, std::wstring const& name)
+{
+	HANDLE file = OpenAt(dir, name, FILE_READ_EA | FILE_READ_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, FILE_OPEN_FOR_BACKUP_INTENT | FILE_OPEN_REPARSE_POINT);
+	if (file != INVALID_HANDLE_VALUE) {
+		FILE_ATTRIBUTE_TAG_INFO buf{};
+
+		bool ret{};
+		if (GetFileInformationByHandleEx(file, FileAttributeTagInfo, &buf, sizeof(FILE_ATTRIBUTE_TAG_INFO)) != 0) {
+			ret = IsReparseTagNameSurrogate(buf.ReparseTag);
+		}
+		CloseHandle(file);
+
+		return ret;
+	}
+
+	return false;
+}
+}
+
+bool local_filesys::check_buffer()
+{
+	if (!cur_) {
+		cur_ = buffer_.data();
+		if (!read_dir_buffer(dir_, cur_, buffer_.size(), lfzFileDirectoryInformation)) {
+			return false;
+		}
+	}
+	return true;
+}
+#endif
+
 bool local_filesys::get_next_file(native_string& name)
 {
 #ifdef FZ_WINDOWS
-	if (!has_next_) {
-		return false;
-	}
-	do {
-		name = m_find_data.cFileName;
-		if (name.empty()) {
-			has_next_ = FindNextFile(m_hFind, &m_find_data) != 0;
-			return true;
+	while (dir_ != INVALID_HANDLE_VALUE) {
+		if (!check_buffer()) {
+			name.clear();
+			end_find_files();
+			return false;
 		}
-		if (name == fzT(".") || name == fzT("..")) {
+
+		auto & data = *reinterpret_cast<lfzFILE_DIRECTORY_INFORMATION*>(cur_);
+		cur_ = data.NextEntryOffset ? (cur_ + data.NextEntryOffset) : nullptr;
+
+		if (!data.FileNameLength || (data.FileNameLength % sizeof(wchar_t))) {
+			continue;
+		}
+		if (dirs_only_ && !(data.FileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
 			continue;
 		}
 
-		if (dirs_only_ && !(m_find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
+		// Array may be larger than neceessary, look for terminating null
+		std::wstring_view v(static_cast<wchar_t const*>(data.FileName), data.FileNameLength / sizeof(wchar_t));
+		size_t pos = v.find(wchar_t(0));
+		name.assign(v.substr(0, pos));
+		if (name.empty() || name == fzT(".") || name == fzT("..")) {
 			continue;
 		}
 
-		has_next_ = FindNextFile(m_hFind, &m_find_data) != 0;
 		return true;
-	} while ((has_next_ = FindNextFile(m_hFind, &m_find_data) != 0));
+	}
 
+	name.clear();
 	return false;
 #else
 	if (!dir_) {
@@ -539,34 +664,42 @@ bool local_filesys::get_next_file(native_string& name)
 bool local_filesys::get_next_file(native_string& name, bool &is_link, local_filesys::type & t, int64_t* size, datetime* modification_time, int* mode)
 {
 #ifdef FZ_WINDOWS
-	if (!has_next_) {
-		return false;
-	}
-	do {
-		if (!m_find_data.cFileName[0]) {
+	while (dir_ != INVALID_HANDLE_VALUE) {
+		if (!check_buffer()) {
+			name.clear();
+			end_find_files();
+			return false;
+		}
+
+		auto& data = *reinterpret_cast<lfzFILE_DIRECTORY_INFORMATION*>(cur_);
+		cur_ = data.NextEntryOffset ? (cur_ + data.NextEntryOffset) : nullptr;
+
+		if (!data.FileNameLength || (data.FileNameLength % sizeof(wchar_t))) {
 			continue;
 		}
-		if (dirs_only_ && !(m_find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
+		if (dirs_only_ && !(data.FileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
 			continue;
 		}
 
-		if (m_find_data.cFileName[0] == '.' && (!m_find_data.cFileName[1] || (m_find_data.cFileName[1] == '.' && !m_find_data.cFileName[2]))) {
+		// Array may be larger than neceessary, look for terminating null
+		std::wstring_view v(static_cast<wchar_t const*>(data.FileName), data.FileNameLength / sizeof(wchar_t));
+		size_t pos = v.find(wchar_t(0));
+		name.assign(v.substr(0, pos));
+		if (name.empty() || name == fzT(".") || name == fzT("..")) {
 			continue;
 		}
-		name = m_find_data.cFileName;
 
-		t = (m_find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? dir : file;
+		t = (data.FileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? dir : file;
 
-		is_link = (m_find_data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0 && IsReparseTagNameSurrogate(m_find_data.dwReserved0);
+		is_link = (data.FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0 && IsNameSurrogateReparsePoint(dir_, name);
 		if (is_link && query_symlink_targets_) {
 			// Follow the reparse point
-			HANDLE hFile = CreateFile((m_find_path + name).c_str(), FILE_READ_ATTRIBUTES | FILE_READ_EA, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, nullptr, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, nullptr);
+			HANDLE hFile = OpenAt(dir_, name, FILE_READ_EA | FILE_READ_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, FILE_OPEN_FOR_BACKUP_INTENT); 
 			if (hFile != INVALID_HANDLE_VALUE) {
 				BY_HANDLE_FILE_INFORMATION info{};
 				int ret = GetFileInformationByHandle(hFile, &info);
 				CloseHandle(hFile);
 				if (ret != 0) {
-
 					if (modification_time) {
 						*modification_time = datetime(info.ftLastWriteTime, datetime::milliseconds);
 						if (modification_time->empty()) {
@@ -588,7 +721,6 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 						}
 					}
 
-					has_next_ = FindNextFile(m_hFind, &m_find_data) != 0;
 					return true;
 				}
 			}
@@ -609,14 +741,19 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 		}
 		else {
 			if (modification_time) {
-				*modification_time = datetime(m_find_data.ftLastWriteTime, datetime::milliseconds);
+				FILETIME ft;
+				ft.dwLowDateTime = data.LastWriteTime.LowPart;
+				ft.dwHighDateTime = data.LastWriteTime.HighPart;
+				*modification_time = datetime(ft, datetime::milliseconds);
 				if (modification_time->empty()) {
-					*modification_time = datetime(m_find_data.ftLastWriteTime, datetime::milliseconds);
+					ft.dwLowDateTime = data.CreationTime.LowPart;
+					ft.dwHighDateTime = data.CreationTime.HighPart;
+					*modification_time = datetime(ft, datetime::milliseconds);
 				}
 			}
 
 			if (mode) {
-				*mode = (int)m_find_data.dwFileAttributes;
+				*mode = (int)data.FileAttributes;
 			}
 
 			if (size) {
@@ -624,13 +761,12 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 					*size = -1;
 				}
 				else {
-					*size = make_int64fzT(m_find_data.nFileSizeHigh, m_find_data.nFileSizeLow);
+					*size = data.EndOfFile.QuadPart;
 				}
 			}
 		}
-		has_next_ = FindNextFile(m_hFind, &m_find_data) != 0;
 		return true;
-	} while ((has_next_ = FindNextFile(m_hFind, &m_find_data) != 0));
+	}
 
 	return false;
 #else
@@ -759,13 +895,17 @@ native_string local_filesys::get_link_target(native_string const& path)
 		CloseHandle(hFile);
 	}
 #else
-	size_t const size = 1024;
-	char out[size];
-
-	ssize_t res = readlink(path.c_str(), out, size);
-	if (res > 0 && static_cast<size_t>(res) < size) {
-		out[res] = 0;
-		target = out;
+	target.resize(1024);
+	while (true) {
+		ssize_t res = readlink(path.c_str(), target.data(), target.size());
+		if (res < 0) {
+			return {};
+		}
+		if (static_cast<size_t>(res) < target.size()) {
+			target.resize(res);
+			break;
+		}
+		target.resize(target.size() * 2);
 	}
 #endif
 	return target;
@@ -774,7 +914,6 @@ native_string local_filesys::get_link_target(native_string const& path)
 namespace {
 result do_mkdir(native_string const& path, mkdir_permissions permissions)
 {
-	result ret{result::other};
 #ifdef FZ_WINDOWS
 
 	SECURITY_ATTRIBUTES attr{};
@@ -795,22 +934,36 @@ result do_mkdir(native_string const& path, mkdir_permissions permissions)
 	}
 	
 	if (CreateDirectory(path.c_str(), &attr)) {
-		ret = {result::ok};
+		return {result::ok};
 	}
-	else if (GetLastError() == ERROR_ACCESS_DENIED) {
-		ret = {result::noperm};
+
+	auto const err = GetLastError();
+	switch (err) {
+	case ERROR_ACCESS_DENIED:
+		return {result::noperm, err};
+	case ERROR_DISK_FULL:
+		return {result::nospace, err};
+	default:
+		return {result::other, err};
 	}
 #else
 	int res = ::mkdir(path.c_str(), (permissions == mkdir_permissions::normal) ? 0777 : 0700);
 	if (!res) {
-		ret = {result::ok};
+		return {result::ok};
 	}
-	else if (errno == EACCES || errno == EPERM) {
-		ret = {result::noperm};
+
+	int const err = errno;
+	switch (err) {
+	case EACCES:
+	case EPERM:
+		return {result::noperm, err};
+	case EDQUOT:
+	case ENOSPC:
+		return {result::nospace, err};
+	default:
+		return {result::other, err};
 	}
 #endif
-
-	return ret;
 }
 }
 
@@ -828,11 +981,11 @@ result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions
 		unc = true;
 		size_t pos = absolute_path.find_first_of(L"\\/", 2);
 		if (pos == std::wstring::npos || pos == 2) {
-			return result{result::other};
+			return {result::invalid};
 		}
 		size_t pos2 = absolute_path.find_first_of(L"\\/", pos + 1);
 		if (pos2 == pos + 1) {
-			return result{result::other};
+			return {result::invalid};
 		}
 		min_len = (pos2 == std::wstring::npos) ? absolute_path.size() : (pos2 - 1);
 	}
@@ -840,17 +993,17 @@ result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions
 	if (!unc) {
 		auto c = absolute_path[offset];
 		if (absolute_path.size() < offset + 2 || absolute_path[offset + 1] != ':' || !is_drive_letter(c)) {
-			return result{result::other};
+			return {result::invalid};
 		}
 		size_t pos = absolute_path.find_first_of(L"\\/", offset + 2);
 		if (pos != std::wstring::npos && pos != offset + 2) {
-			return result{result::other};
+			return result{result::invalid};
 		}
 		min_len = offset + 2;
 	}
 #else
 	if (absolute_path.empty() || absolute_path[0] != '/') {
-		return result{result::other};
+		return {result::invalid};
 	}
 	size_t const min_len = 1;
 #endif
@@ -893,7 +1046,7 @@ result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions
 					break;
 				}
 				else if (t != fz::local_filesys::unknown) {
-					return result{result::nodir};
+					return {result::nodir};
 				}
 			}
 			else {
@@ -901,7 +1054,7 @@ result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions
 			}
 		}
 		if (!found) {
-			return result{result::other};
+			return {result::other};
 		}
 
 		// Step 3: Create the segments
@@ -927,7 +1080,42 @@ result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions
 		}
 	}
 
-	return result{result::ok};
+	return {result::ok};
+}
+
+result remove_dir(native_string const& absolute_path)
+{
+	if (absolute_path.empty()) {
+		return {result::invalid};
+	}
+#if FZ_WINDOWS
+	if (!RemoveDirectoryW(absolute_path.c_str())) {
+		DWORD const err = GetLastError();
+		switch (err) {
+			case ERROR_PATH_NOT_FOUND:
+				return {result::nodir, err};
+			case ERROR_ACCESS_DENIED:
+				return {result::noperm, err};
+			default:
+				return {result::other, err};
+		}
+	}
+#else
+	if (rmdir(absolute_path.c_str()) != 0) {
+		int const err = errno;
+		switch (err) {
+			case EACCES:
+			case EPERM:
+				return {result::noperm, err};
+			case ENOTDIR:
+			case ENOENT:
+				return {result::nodir, err};
+			default:
+				return {result::other, err};
+		}
+	}
+#endif
+	return {result::ok};
 }
 
 #ifndef FZ_WINDOWS
@@ -993,18 +1181,18 @@ result rename_file(native_string const& source, native_string const& dest, bool
 		return {result::ok};
 	}
 
-	DWORD err = GetLastError();
+	DWORD const err = GetLastError();
 	switch (err) {
 		case ERROR_FILE_NOT_FOUND:
-			return {result::nofile};
+			return {result::nofile, err};
 		case ERROR_PATH_NOT_FOUND:
-		return {result::nodir};
+		return {result::nodir, err};
 		case ERROR_ACCESS_DENIED:
-			return {result::noperm};
+			return {result::noperm, err};
 		case ERROR_DISK_FULL:
-			return {result::nospace};
+			return {result::nospace, err};
 		default:
-			return {result::other};
+			return {result::other, err};
 	}
 #else
 	int res = rename(source.c_str(), dest.c_str());
@@ -1012,25 +1200,25 @@ result rename_file(native_string const& source, native_string const& dest, bool
 		return {result::ok};
 	}
 
-	switch (errno) {
+	int const err = errno;
+	switch (err) {
 	case EPERM:
 	case EACCES:
-		return {result::noperm};
+		return {result::noperm, err};
 	case ENOSPC:
-		return {result::nospace};
+		return {result::nospace, err};
 	case ENOTDIR:
-		return {result::nodir};
+		return {result::nodir, err};
 	case ENOENT:
 	case EISDIR:
-		return {result::nofile};
+		return {result::nofile, err};
 	case EXDEV:
+		if (!allow_copy) {
+			return {result::other, err};
+		}
 		break;
 	default:
-		return {result::other};
-	}
-
-	if (!allow_copy) {
-		return {result::other};
+		return {result::other, err};
 	}
 
 	bool dest_opened{};
@@ -1044,19 +1232,18 @@ result rename_file(native_string const& source, native_string const& dest, bool
 
 	res = unlink(source.c_str());
 	if (res != 0) {
-		switch (errno) {
+		int const err = errno;
+		switch (err) {
 		case EPERM:
 		case EACCES:
-			return {result::noperm};
+			return {result::noperm, err};
 		case ENOTDIR:
-			return {result::nodir};
+			return {result::nodir, err};
 		case ENOENT:
 		case EISDIR:
-			return {result::nofile};
-		case EXDEV:
-			break;
+			return {result::nofile, err};
 		default:
-			return {result::other};
+			return {result::other, err};
 		}
 	}
 
@@ -1064,4 +1251,59 @@ result rename_file(native_string const& source, native_string const& dest, bool
 #endif
 }
 
+local_filesys::local_filesys(local_filesys && op)
+{
+#if FZ_WINDOWS
+	uintptr_t offset{};
+	if (op.cur_) {
+		offset = op.cur_ - op.buffer_.data();
+	}
+
+	buffer_ = std::move(op.buffer_);
+	if (op.cur_) {
+		cur_ = buffer_.data() + offset;
+	}
+	dir_ = op.dir_;
+
+	op.buffer_.clear();
+	op.cur_ = nullptr;
+	op.dir_ = INVALID_HANDLE_VALUE;
+#else
+	dir_ = op.dir_;
+	op.dir_ = nullptr;
+#endif
+	dirs_only_ = op.dirs_only_;
+	query_symlink_targets_ = op.query_symlink_targets_;
+}
+
+local_filesys& local_filesys::operator=(local_filesys && op)
+{
+	if (&op != this) {
+		end_find_files();
+
+#if FZ_WINDOWS
+		uintptr_t offset{};
+		if (op.cur_) {
+			offset = op.cur_ - op.buffer_.data();
+		}
+
+		buffer_ = std::move(op.buffer_);
+		if (op.cur_) {
+			cur_ = buffer_.data() + offset;
+		}
+		dir_ = op.dir_;
+
+		op.buffer_.clear();
+		op.cur_ = nullptr;
+		op.dir_ = INVALID_HANDLE_VALUE;
+#else
+		dir_ = op.dir_;
+		op.dir_ = nullptr;
+#endif
+		dirs_only_ = op.dirs_only_;
+		query_symlink_targets_ = op.query_symlink_targets_;
+	}
+	return *this;
+}
+
 }
