@@ -156,7 +156,7 @@ public:
 			err_.create(true);
 	}
 
-	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io, impersonation_token const* it = nullptr)
+	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, io_redirection redirect_mode, impersonation_token const* it = nullptr)
 	{
 		if (process_ != INVALID_HANDLE_VALUE) {
 			return false;
@@ -165,7 +165,7 @@ public:
 		STARTUPINFO si{};
 		si.cb = sizeof(si);
 
-		if (redirect_io) {
+		if (redirect_mode != io_redirection::none) {
 			if (!create_pipes()) {
 				return false;
 			}
@@ -197,11 +197,16 @@ public:
 		process_ = pi.hProcess;
 
 		// We don't need to use these
-		if (redirect_io) {
+		if (redirect_mode != io_redirection::none) {
 			reset_handle(pi.hThread);
 			reset_handle(in_.read_);
 			reset_handle(out_.write_);
 			reset_handle(err_.write_);
+			if (redirect_mode == io_redirection::closeall) {
+				reset_handle(in_.write_);
+				reset_handle(out_.read_);
+				reset_handle(err_.read_);
+			}
 		}
 
 		return true;
@@ -263,6 +268,7 @@ private:
 #else
 
 #include "libfilezilla/glue/unix.hpp"
+#include "libfilezilla/mutex.hpp"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -271,6 +277,7 @@ private:
 #include <string.h>
 #include <unistd.h>
 
+#include <atomic>
 #include <memory>
 #include <vector>
 
@@ -308,12 +315,12 @@ public:
 	pipe(pipe const&) = delete;
 	pipe& operator=(pipe const&) = delete;
 
-	bool create(bool require_atomic_creation = false)
+	bool create()
 	{
 		reset();
 
 		int fds[2];
-		if (!create_pipe(fds, require_atomic_creation)) {
+		if (!create_pipe(fds)) {
 			return false;
 		}
 
@@ -337,28 +344,18 @@ public:
 	int write_{-1};
 };
 
-void make_arg(native_string const& arg, std::vector<std::unique_ptr<native_string::value_type[]>> & argList)
+void get_argv(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, std::vector<char*> & argV)
 {
-	auto ret = std::make_unique<char[]>(arg.size() + 1);
-	memcpy(ret.get(), arg.c_str(), arg.size() + 1);
-	argList.push_back(std::move(ret));
-}
-
-void get_argv(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, std::vector<std::unique_ptr<char[]>> & argList, std::unique_ptr<char *[]> & argV)
-{
-	argList.reserve(end - begin + 1);
-	make_arg(cmd, argList);
+	argV.reserve(end - begin + 2);
+	argV.emplace_back(const_cast<char*>(cmd.c_str()));
 	for (auto it = begin; it != end; ++it) {
-		make_arg(*it, argList);
+		argV.emplace_back(const_cast<char*>(it->c_str()));
 	}
-
-	argV = std::make_unique<char*[]>(argList.size() + 1);
-	char ** v = argV.get();
-	for (auto const& a : argList) {
-		*(v++) = a.get();
-	}
-	*v = nullptr;
+	argV.emplace_back(nullptr);
 }
+
+std::atomic<unsigned int> forkblocks_{};
+mutex forkblock_mtx_;
 }
 
 class process::impl
@@ -381,20 +378,20 @@ public:
 			err_.create();
 	}
 
-	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io, std::vector<int> const& extra_fds = std::vector<int>(), impersonation_token const* it = nullptr)
+	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, io_redirection redirect_mode, std::vector<int> const& extra_fds = std::vector<int>(), impersonation_token const* it = nullptr)
 	{
 		if (pid_ != -1) {
 			return false;
 		}
 
-		if (redirect_io && !create_pipes()) {
+		if (redirect_mode != io_redirection::none && !create_pipes()) {
 			return false;
 		}
 
-		std::vector<std::unique_ptr<char[]>> argList;
-		std::unique_ptr<char *[]> argV;
-		get_argv(cmd, begin, end, argList, argV);
+		std::vector<char*> argV;
+		get_argv(cmd, begin, end, argV);
 
+		scoped_lock l(forkblock_mtx_);
 		pid_t pid = fork();
 		if (pid < 0) {
 			return false;
@@ -402,7 +399,7 @@ public:
 		else if (!pid) {
 			// We're the child.
 
-			if (redirect_io) {
+			if (redirect_mode != io_redirection::none) {
 				// Close uneeded descriptors
 				reset_fd(in_.write_);
 				reset_fd(out_.read_);
@@ -410,6 +407,8 @@ public:
 
 				// Redirect to pipe. The redirected descriptors don't have
 				// FD_CLOEXEC set.
+				// Note that even if redirect_mode is closeall, we still leave valid descriptors
+				// at stdin/out/err as we do not want to have these re-used for other things.
 				if (dup2(in_.read_, STDIN_FILENO) == -1 ||
 					dup2(out_.write_, STDOUT_FILENO) == -1 ||
 					dup2(err_.write_, STDERR_FILENO) == -1)
@@ -429,16 +428,14 @@ public:
 				}
 			}
 
-#if FZ_UNIX
 			if (it && *it) {
 				if (!set_process_impersonation(*it)) {
 					_exit(1);
 				}
 			}
-#endif
 
 			// Execute process
-			execv(cmd.c_str(), argV.get()); // noreturn on success
+			execv(cmd.c_str(), argV.data()); // noreturn on success
 
 			_exit(-1);
 		}
@@ -447,10 +444,15 @@ public:
 			pid_ = pid;
 
 			// Close unneeded descriptors
-			if (redirect_io) {
+			if (redirect_mode != io_redirection::none) {
 				reset_fd(in_.read_);
 				reset_fd(out_.write_);
 				reset_fd(err_.write_);
+				if (redirect_mode == io_redirection::closeall) {
+					reset_fd(in_.write_);
+					reset_fd(out_.read_);
+					reset_fd(err_.read_);
+				}
 			}
 		}
 
@@ -525,42 +527,42 @@ process::~process()
 	delete impl_;
 }
 
-bool process::spawn(native_string const& cmd, std::vector<native_string> const& args, bool redirect_io)
+bool process::spawn(native_string const& cmd, std::vector<native_string> const& args, io_redirection redirect_mode)
 {
-	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io) : false;
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_mode) : false;
 }
 
 #if FZ_WINDOWS || FZ_UNIX
-bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, bool redirect_io)
+bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, io_redirection redirect_mode)
 {
 #if FZ_WINDOWS
-	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, &it) : false;
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_mode, &it) : false;
 #else
-	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, {}, &it) : false;
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_mode, {}, &it) : false;
 #endif
 }
 #endif
 
 #ifndef FZ_WINDOWS
-bool process::spawn(native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, bool redirect_io)
+bool process::spawn(native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, io_redirection redirect_mode)
 {
-	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, extra_fds) : false;
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_mode, extra_fds) : false;
 }
 
-bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, bool redirect_io)
+bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, io_redirection redirect_mode)
 {
-	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, extra_fds, &it) : false;
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_mode, extra_fds, &it) : false;
 }
 
 #endif
 
-bool process::spawn(std::vector<native_string> const& command_with_args, bool redirect_io)
+bool process::spawn(std::vector<native_string> const& command_with_args, io_redirection redirect_mode)
 {
 	if (command_with_args.empty()) {
 		return false;
 	}
 	auto begin = command_with_args.begin() + 1;
-	return impl_ ? impl_->spawn(command_with_args.front(), begin, command_with_args.end(), redirect_io) : false;
+	return impl_ ? impl_->spawn(command_with_args.front(), begin, command_with_args.end(), redirect_mode) : false;
 }
 
 void process::kill()
@@ -767,20 +769,21 @@ bool spawn_detached_process(std::vector<native_string> const& cmd_with_args)
 	}
 #endif
 
-	std::vector<std::unique_ptr<char[]>> argList;
-	std::unique_ptr<char *[]> argV;
+	std::vector<char *> argV;
 	auto begin = cmd_with_args.cbegin() + 1;
-	get_argv(cmd_with_args.front(), begin, cmd_with_args.cend(), argList, argV);
+	get_argv(cmd_with_args.front(), begin, cmd_with_args.cend(), argV);
 
 	pid_t const parent = getppid();
 	pid_t const ppgid = getpgid(parent);
 
+	scoped_lock l(forkblock_mtx_);
+
 	// We're using a pipe created with O_CLOEXEC to signal failure from execv.
-	// Note that this flags must be set atomically, setting FD_CLOEXEC after creation
-	// leads to a deadlock if a different thread calls exec in-between.
-	// Unfortunately even in early 2020, macOS does not have pipe2.
+	// Fortunately the forkblock avoids a deadlock if the cloexec flag isn't set
+	// atomically and another threa execs in-between, as even in 2022, macOS
+	// doesn't have pipe2.
 	pipe errpipe;
-	errpipe.create(true);
+	errpipe.create();
 
 	pid_t pid = fork();
 	if (!pid) {
@@ -791,7 +794,7 @@ bool spawn_detached_process(std::vector<native_string> const& cmd_with_args)
 		if (!inner_pid) {
 			// Change the process group ID of the new process so that terminating the outer process does not terminate the child
 			setpgid(0, ppgid);
-			execv(argV.get()[0], argV.get());
+			execv(cmd_with_args[0].c_str(), argV.data());
 
 			if (errpipe.write_ != -1) {
 				ssize_t w;
@@ -836,4 +839,52 @@ bool spawn_detached_process(std::vector<native_string> const& cmd_with_args)
 	}
 #endif
 }
+
+#if !FZ_WINDOWS
+forkblock::forkblock()
+{
+	forkblock_mtx_.lock();
+	++forkblocks_;
+}
+
+forkblock::~forkblock()
+{
+	--forkblocks_;
+	forkblock_mtx_.unlock();
+}
+
+namespace {
+bool forked_child_{};
+void atfork_lock_forkblock()
+{
+	// Unsafe if fork is called from a signal handler
+	if (!forked_child_) {
+		forkblock_mtx_.lock();
+	}
+}
+
+void atfork_unlock_forkblock()
+{
+	// Unsafe if fork is called from a signal handler
+	if (!forked_child_) {
+		forkblock_mtx_.unlock();
+	}
+}
+
+void atfork_check_forkblocks()
+{
+	// Last line of defense.
+	if (forkblocks_) {
+		_exit(1);
+	}
+	forked_child_ = true;
+}
+
+
+int const atfork_registered = []() {
+	return pthread_atfork(&atfork_lock_forkblock, &atfork_unlock_forkblock, &atfork_check_forkblocks);
+}();
+}
+#endif
+
 }
