@@ -1,7 +1,10 @@
-#include "libfilezilla/recursive_remove.hpp"
+#include "libfilezilla/file.hpp"
 #include "libfilezilla/local_filesys.hpp"
+#include "libfilezilla/recursive_remove.hpp"
 
-#ifndef FZ_WINDOWS
+#if FZ_WINDOWS
+#include "windows/dll.hpp"
+#else
 #include <unistd.h>
 #endif
 
@@ -14,6 +17,12 @@ bool recursive_remove::remove(const native_string& path)
 	return remove(paths);
 }
 
+#if FZ_WINDOWS
+extern "C" {
+typedef int (*shfileop_t)(LPSHFILEOPSTRUCTW);
+}
+#endif
+
 bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 {
 	bool success = true;
@@ -23,46 +32,51 @@ bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 	// to delete all contents.
 
 #ifdef FZ_WINDOWS
-	// SHFileOperation accepts a list of null-terminated strings. Go through all
-	// paths to get the required buffer length
+	shdlls& dlls = shdlls::get();
+	static shfileop_t const shfileop = dlls.shell32_ ? reinterpret_cast<shfileop_t>(GetProcAddress(dlls.shell32_.h_, "SHFileOperationW")) : nullptr;
+	if (shfileop) {
+		// SHFileOperation accepts a list of null-terminated strings. Go through all
+		// paths to get the required buffer length
 
-	size_t len = 1; // String list terminated by empty string
+		size_t len = 1; // String list terminated by empty string
 
-	for (auto const& dir : dirsToVisit) {
-		len += dir.size() + 1;
-	}
-
-	// Allocate memory
-	native_string::value_type* pBuffer = new native_string::value_type[len];
-	native_string::value_type* p = pBuffer;
-
-	for (auto& dir : dirsToVisit) {
-		if (!dir.empty() && local_filesys::is_separator(dir.back())) {
-			dir.pop_back();
-		}
-		if (local_filesys::get_file_type(dir) == local_filesys::unknown) {
-			continue;
+		for (auto const& dir : dirsToVisit) {
+			len += dir.size() + 1;
 		}
 
-		wcscpy(p, dir.c_str());
-		p += dir.size() + 1;
-	}
-	if (p != pBuffer) {
-		*p = 0;
+		// Allocate memory
+		native_string::value_type* pBuffer = new native_string::value_type[len];
+		native_string::value_type* p = pBuffer;
 
-		// Now we can delete the files in the buffer
-		SHFILEOPSTRUCT op{};
-		op.wFunc = FO_DELETE;
-		op.pFrom = pBuffer;
+		for (auto& dir : dirsToVisit) {
+			if (!dir.empty() && local_filesys::is_separator(dir.back())) {
+				dir.pop_back();
+			}
+			if (local_filesys::get_file_type(dir) == local_filesys::unknown) {
+				continue;
+			}
 
-		adjust_shfileop(op);
-
-		if (SHFileOperation(&op) != 0) {
-			success = false;
+			wcscpy(p, dir.c_str());
+			p += dir.size() + 1;
 		}
+		if (p != pBuffer) {
+			*p = 0;
+
+			// Now we can delete the files in the buffer
+			SHFILEOPSTRUCTW op{};
+			op.wFunc = FO_DELETE;
+			op.pFrom = pBuffer;
+
+			adjust_shfileop(op);
+
+			if (shfileop(&op) != 0) {
+				success = false;
+			}
+		}
+		delete [] pBuffer;
+		return success;
 	}
-	delete [] pBuffer;
-#else
+#endif
 	if (!confirm()) {
 		return false;
 	}
@@ -78,7 +92,7 @@ bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 
 	local_filesys fs;
 
-	// Process all dirctories that have to be visited
+	// Process all directories that have to be visited
 	while (!dirsToVisit.empty()) {
 		auto const iter = dirsToVisit.begin();
 		native_string const& path = *iter;
@@ -89,7 +103,7 @@ bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 		}
 
 		if (fs.get_file_type(path) != local_filesys::dir) {
-			if (unlink(path.c_str()) != 0) {
+			if (!remove_file(path)) {
 				success = false;
 			}
 			dirsToVisit.erase(iter);
@@ -127,7 +141,7 @@ bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 
 		// Delete all files and links in current directory enumerated before
 		for (auto const& filename : filesToDelete) {
-			if (unlink(filename.c_str()) != 0) {
+			if (!remove_file(filename)) {
 				success = false;
 			}
 		}
@@ -135,11 +149,10 @@ bool recursive_remove::remove(std::list<native_string> dirsToVisit)
 
 	// Delete the now empty directories
 	for (auto const& dir : dirsToDelete) {
-		if (rmdir(dir.c_str()) != 0) {
+		if (!remove_dir(dir)) {
 			success = false;
 		}
 	}
-#endif
 
 	return success;
 }
