@@ -15,6 +15,8 @@
 
 #include <string.h>
 
+using namespace std::literals;
+
 #if DEBUG_SOCKETEVENTS
 #include <assert.h>
 
@@ -23,8 +25,6 @@ bool FZ_PRIVATE_SYMBOL has_pending_event(event_handler * handler, socket_event_s
 }
 #endif
 
-static_assert(GNUTLS_VERSION_NUMBER != 0x030604, "Using TLS 1.3 with this version of GnuTLS does not work, update your version of GnuTLS");
-
 namespace fz {
 
 namespace {
@@ -32,11 +32,7 @@ namespace {
 #if FZ_USE_GNUTLS_SYSTEM_CIPHERS
 char const ciphers[] = "@SYSTEM:-ARCFOUR-128:-3DES-CBC:-MD5:-SIGN-RSA-MD5:-VERS-SSL3.0";
 #else
-	#if GNUTLS_VERSION_NUMBER >= 0x030600
-		char const ciphers[] = "SECURE256:+SECURE128:-ARCFOUR-128:-3DES-CBC:-MD5:+SIGN-ALL:-SIGN-RSA-MD5:+CTYPE-X509:-VERS-SSL3.0";
-	#else
-		char const ciphers[] = "SECURE256:+SECURE128:-ARCFOUR-128:-3DES-CBC:-MD5:+SIGN-ALL:-SIGN-RSA-MD5:+CTYPE-X509:-CTYPE-OPENPGP:-VERS-SSL3.0";
-	#endif
+	char const ciphers[] = "SECURE256:+SECURE128:-ARCFOUR-128:-3DES-CBC:-MD5:+SIGN-ALL:-SIGN-RSA-MD5:+CTYPE-X509:-VERS-SSL3.0";
 #endif
 
 #define TLSDEBUG 0
@@ -117,11 +113,60 @@ public:
 
 		char const* name = gnutls_handshake_description_get_name(static_cast<gnutls_handshake_description_t>(htype));
 
-		tls->logger_.log(logmsg::debug_debug, L"TLS handshake: %s %s", prefix, name);
+		tls->logger_.log(logmsg::debug_debug, L"TLS handshakep: %s %s", prefix, name);
 
 		return 0;
 	}
+
+	static int store_session(void* ptr, gnutls_datum_t const& key, gnutls_datum_t const& data)
+	{
+		auto* tls = reinterpret_cast<tls_layer_impl*>(ptr);
+		if (!tls) {
+			return 0;
+		}
+		if (!key.size || !data.size) {
+			return 0;
+		}
+		tls->session_db_key_.resize(key.size);
+		memcpy(tls->session_db_key_.data(), key.data, key.size);
+		tls->session_db_data_.resize(data.size);
+		memcpy(tls->session_db_data_.data(), data.data, data.size);
+
+		return 0;
+	}
+
+	static gnutls_datum_t retrieve_session(void *ptr, gnutls_datum_t key)
+	{
+		auto* tls = reinterpret_cast<tls_layer_impl*>(ptr);
+		if (!tls) {
+			return {};
+		}
+		if (!key.size) {
+			return {};
+		}
+
+		if (key.size == tls->session_db_key_.size() && !memcmp(tls->session_db_key_.data(), key.data, key.size)) {
+			gnutls_datum_t d{};
+			d.data = reinterpret_cast<unsigned char*>(gnutls_malloc(tls->session_db_data_.size()));
+			if (d.data) {
+				d.size = tls->session_db_data_.size();
+				memcpy(d.data, tls->session_db_data_.data(), d.size);
+			}
+			return d;
+		}
+
+		return gnutls_datum_t{};
+	}
+
+	static void verify_output_cb(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)
+	{
+		if (verify_output_cb_) {
+			verify_output_cb_(cert, issuer, crl, verification_output);
+		}
+	}
+	static thread_local std::function<void(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)> verify_output_cb_;
 };
+thread_local std::function<void(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)> tls_layerCallbacks::verify_output_cb_;
 
 namespace {
 extern "C" int handshake_hook_func(gnutls_session_t session, unsigned int htype, unsigned int post, unsigned int incoming, gnutls_datum_t const*)
@@ -129,22 +174,20 @@ extern "C" int handshake_hook_func(gnutls_session_t session, unsigned int htype,
 	return tls_layerCallbacks::handshake_hook_func(session, htype, post, incoming);
 }
 
-struct cert_list_holder final
+extern "C" int db_store_func(void *ptr, gnutls_datum_t key, gnutls_datum_t data)
 {
-	cert_list_holder() = default;
-	~cert_list_holder() {
-		for (unsigned int i = 0; i < certs_size; ++i) {
-			gnutls_x509_crt_deinit(certs[i]);
-		}
-		gnutls_free(certs);
-	}
+	return tls_layerCallbacks::store_session(ptr, key, data);
+}
 
-	cert_list_holder(cert_list_holder const&) = delete;
-	cert_list_holder& operator=(cert_list_holder const&) = delete;
-
-	gnutls_x509_crt_t * certs{};
-	unsigned int certs_size{};
-};
+extern "C" gnutls_datum_t db_retr_func(void *ptr, gnutls_datum_t key)
+{
+	return tls_layerCallbacks::retrieve_session(ptr, key);
+}
+extern "C" int c_verify_output_cb(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)
+{
+	tls_layerCallbacks::verify_output_cb(cert, issuer, crl, verification_output);
+	return 0;
+}
 
 std::string to_string(gnutls_datum_t const& d)
 {
@@ -154,6 +197,14 @@ std::string to_string(gnutls_datum_t const& d)
 	return {};
 }
 
+std::string_view to_view(gnutls_datum_t const& d)
+{
+	if (d.data && d.size) {
+		return std::string_view(reinterpret_cast<char const*>(d.data), d.size);
+	}
+	return {};
+}
+
 struct datum_holder final : gnutls_datum_t
 {
 	datum_holder() {
@@ -247,6 +298,34 @@ bool tls_layer_impl::init()
 	return true;
 }
 
+std::string read_certificates_file(native_string const& certsfile, logger_interface * logger)
+{
+	file cf(certsfile, file::reading, file::existing);
+	if (!cf.opened()) {
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Could not open certificate file."));
+		}
+		return {};
+	}
+	int64_t const cs = cf.size();
+	if (cs < 0 || cs > 1024 * 1024) {
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Certificate file too big."));
+		}
+		return {};
+	}
+	std::string c;
+	c.resize(cs);
+	auto read = cf.read(c.data(), cs);
+	if (read != cs) {
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Could not read certificate file."));
+		}
+		return {};
+	}
+	return c;
+}
+
 bool tls_layer_impl::set_certificate_file(native_string const& keyfile, native_string const& certsfile, native_string const& password, bool pem)
 {
 	// Load the files ourselves instead of calling gnutls_certificate_set_x509_key_file2
@@ -270,29 +349,15 @@ bool tls_layer_impl::set_certificate_file(native_string const& keyfile, native_s
 		return false;
 	}
 
-
-	file cf(certsfile, file::reading, file::existing);
-	if (!cf.opened()) {
-		logger_.log(logmsg::error, fztranslate("Could not open certificate file."));
-		return false;
-	}
-	int64_t const cs = cf.size();
-	if (cs < 0 || cs > 1024 * 1024) {
-		logger_.log(logmsg::error, fztranslate("Certificate file too big."));
-		return false;
-	}
-	std::string c;
-	c.resize(cs);
-	read = cf.read(c.data(), cs);
-	if (read != cs) {
-		logger_.log(logmsg::error, fztranslate("Could not read certificate file."));
+	std::string c = read_certificates_file(certsfile, &logger_);
+	if (c.empty()) {
 		return false;
 	}
 
 	return set_certificate(k, c, password, pem);
 }
 
-bool tls_layer_impl::set_certificate(std::string const& key, std::string const& certs, native_string const& password, bool pem)
+bool tls_layer_impl::set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem)
 {
 	if (!init()) {
 		return false;
@@ -321,14 +386,16 @@ bool tls_layer_impl::set_certificate(std::string const& key, std::string const&
 	return true;
 }
 
-bool tls_layer_impl::init_session(bool client)
+bool tls_layer_impl::init_session(bool client, int extra_flags)
 {
 	if (!cert_credentials_) {
 		deinit();
 		return false;
 	}
 
-	int res = gnutls_init(&session_, client ? GNUTLS_CLIENT : GNUTLS_SERVER);
+	int flags = client ? GNUTLS_CLIENT : GNUTLS_SERVER;
+	flags |= extra_flags;
+	int res = gnutls_init(&session_, flags);
 	if (res) {
 		log_error(res, L"gnutls_init");
 		deinit();
@@ -360,6 +427,7 @@ bool tls_layer_impl::init_session(bool client)
 
 	// For use in callbacks
 	gnutls_session_set_ptr(session_, this);
+	gnutls_db_set_ptr(session_, this);
 
 	// Even though the name gnutls_db_set_cache_expiration
 	// implies expiration of some cache, it also governs
@@ -367,7 +435,44 @@ bool tls_layer_impl::init_session(bool client)
 	// session is cached or not.
 	gnutls_db_set_cache_expiration(session_, 100000000);
 
-	res = gnutls_priority_set_direct(session_, ciphers, nullptr);
+	if (!client) {
+		gnutls_db_set_ptr(session_, this);
+		gnutls_db_set_store_function(session_, &db_store_func);
+		gnutls_db_set_retrieve_function(session_, &db_retr_func);
+	}
+
+	std::string prio = ciphers;
+	switch (min_tls_ver_) {
+	case tls_ver::v1_3:
+		prio += ":-VERS-TLS1.2";
+		// Fallthrough
+	case tls_ver::v1_2:
+		prio += ":-VERS-TLS1.1";
+		// Fallthrough
+	case tls_ver::v1_1:
+		prio += ":-VERS-TLS1.0";
+		break;
+	default:
+		break;
+	}
+
+	if (max_tls_ver_) {
+		switch (*max_tls_ver_) {
+		case tls_ver::v1_0:
+			prio += ":-VERS-TLS1.1";
+			// Fallthrough
+		case tls_ver::v1_1:
+			prio += ":-VERS-TLS1.2";
+			// Fallthrough
+		case tls_ver::v1_2:
+			prio += ":-VERS-TLS1.3";
+			break;
+		default:
+			break;
+		}
+	}
+
+	res = gnutls_priority_set_direct(session_, prio.c_str(), nullptr);
 	if (res) {
 		log_error(res, L"gnutls_priority_set_direct");
 		deinit();
@@ -614,6 +719,9 @@ void tls_layer_impl::on_read()
 {
 	logger_.log(logmsg::debug_debug, L"tls_layer_impl::on_read()");
 
+#if DEBUG_SOCKETEVENTS
+	assert(!can_read_from_socket_);
+#endif
 	can_read_from_socket_ = true;
 
 	if (!session_) {
@@ -624,11 +732,11 @@ void tls_layer_impl::on_read()
 		continue_handshake();
 	}
 	else if (state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) {
-		if (tls_layer_.event_handler_) {
 #if DEBUG_SOCKETEVENTS
-			assert(!debug_can_read_);
-			debug_can_read_ = true;
+		assert(!debug_can_read_);
+		debug_can_read_ = true;
 #endif
+		if (tls_layer_.event_handler_) {
 			tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 		}
 	}
@@ -667,11 +775,7 @@ void tls_layer_impl::on_send()
 
 int tls_layer_impl::continue_write()
 {
-	if (send_buffer_.empty()) {
-		return 0;
-	}
-
-	do {
+	while (!send_buffer_.empty()) {
 		ssize_t res = GNUTLS_E_AGAIN;
 		while ((res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) && can_write_to_socket_) {
 			res = gnutls_record_send(session_, send_buffer_.get(), send_buffer_.size());
@@ -694,17 +798,34 @@ int tls_layer_impl::continue_write()
 
 		send_buffer_.consume(static_cast<size_t>(res));
 	}
-	while (!send_buffer_.empty());
+
+	if (send_new_ticket_) {
+		int res = GNUTLS_E_AGAIN;
+		while ((res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) && can_write_to_socket_) {
+			res = gnutls_session_ticket_send(session_, 1, 0);
+		}
+
+		if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) {
+			return EAGAIN;
+		}
+
+		if (res < 0) {
+			failure(static_cast<int>(res), true);
+			return ECONNABORTED;
+		}
+
+		send_new_ticket_ = false;
+	}
 
 	if (write_blocked_by_send_buffer_) {
 		write_blocked_by_send_buffer_ = false;
 
 		if (state_ == socket_state::connected) {
-			if (tls_layer_.event_handler_) {
 #if DEBUG_SOCKETEVENTS
-				assert(!debug_can_write_);
-				debug_can_write_ = true;
+			assert(!debug_can_write_);
+			debug_can_write_ = true;
 #endif
+			if (tls_layer_.event_handler_) {
 				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
 			}
 		}
@@ -727,6 +848,8 @@ bool tls_layer_impl::client_handshake(std::vector<uint8_t> const& session_to_res
 		return false;
 	}
 
+	server_ = false;
+
 	if (!init() || !init_session(true)) {
 		return false;
 	}
@@ -796,7 +919,28 @@ bool tls_layer_impl::client_handshake(std::vector<uint8_t> const& session_to_res
 	return continue_handshake() == EAGAIN;
 }
 
-bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble)
+namespace {
+bool extract_with_size(uint8_t const* &p, uint8_t const* const end, std::vector<uint8_t>& target)
+{
+	size_t s;
+	if (static_cast<size_t>(end - p) < sizeof(s)) {
+		return false;
+	}
+	memcpy(&s, p, sizeof(s));
+	p += sizeof(s);
+	if (static_cast<size_t>(end - p) < s) {
+		return false;
+	}
+	target.resize(s);
+	if (s) {
+		memcpy(target.data(), p, s);
+		p += s;
+	}
+	return true;
+}
+}
+
+bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble, tls_server_flags flags)
 {
 	logger_.log(logmsg::debug_verbose, L"tls_layer_impl::server_handshake()");
 
@@ -807,9 +951,25 @@ bool tls_layer_impl::server_handshake(std::vector<uint8_t> const& session_to_res
 
 	server_ = true;
 
-	ticket_key_ = session_to_resume;
+	if (!session_to_resume.empty()) {
+		auto const* p = session_to_resume.data();
+		auto const* const end = p + session_to_resume.size();
+		if (!extract_with_size(p, end, ticket_key_)) {
+			return false;
+		}
+		if (!extract_with_size(p, end, session_db_key_)) {
+			return false;
+		}
+		if (!extract_with_size(p, end, session_db_data_)) {
+			return false;
+		}
+	}
 
-	if (!init() || !init_session(false)) {
+	int extra_flags{};
+	if (flags & tls_server_flags::no_auto_ticket) {
+		extra_flags |= GNUTLS_NO_AUTO_SEND_TICKET;
+	}
+	if (!init() || !init_session(false, extra_flags)) {
 		return false;
 	}
 
@@ -878,27 +1038,27 @@ int tls_layer_impl::continue_handshake()
 		std::string const cipherName = get_cipher();
 		std::string const macName = get_mac();
 
-		logger_.log(logmsg::debug_info, L"Protocol: %s, Key exchange: %s, Cipher: %s, MAC: %s", protocol, keyExchange, cipherName, macName);
+		logger_.log(logmsg::debug_info, L"Protocol: %s, Key exchange: %s, Cipher: %s, MAC: %s, ALPN: %s", protocol, keyExchange, cipherName, macName, get_alpn());
 
-		if (is_client()) {
+		if (!server_) {
 			return verify_certificate();
 		}
 		else {
 			state_ = socket_state::connected;
 
+#if DEBUG_SOCKETEVENTS
+			if (can_read_from_socket_) {
+				assert(!debug_can_read_);
+				debug_can_read_ = true;
+			}
+			assert(!debug_can_write_);
+			debug_can_write_ = true;
+#endif
 			if (tls_layer_.event_handler_) {
 				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::connection, 0);
 				if (can_read_from_socket_) {
-#if DEBUG_SOCKETEVENTS
-					assert(!debug_can_read_);
-					debug_can_read_ = true;
-#endif
 					tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 				}
-#if DEBUG_SOCKETEVENTS
-				assert(!debug_can_write_);
-				debug_can_write_ = true;
-#endif
 			}
 		}
 
@@ -956,6 +1116,7 @@ int tls_layer_impl::read(void *buffer, unsigned int len, int& error)
 
 int tls_layer_impl::write(void const* buffer, unsigned int len, int& error)
 {
+//	for(size_t i = 0; i < 20; ++i) {logger_.log(logmsg::error, "Why not Zoidberg?");}
 	if (state_ == socket_state::connecting) {
 		error = EAGAIN;
 		return -1;
@@ -974,7 +1135,7 @@ int tls_layer_impl::write(void const* buffer, unsigned int len, int& error)
 	assert(!has_pending_event(tls_layer_.event_handler_, &tls_layer_, socket_event_flag::write));
 #endif
 
-	if (!send_buffer_.empty()) {
+	if (!send_buffer_.empty() || send_new_ticket_) {
 		write_blocked_by_send_buffer_ = true;
 #if DEBUG_SOCKETEVENTS
 		debug_can_write_ = false;
@@ -1021,18 +1182,20 @@ void tls_layer_impl::failure(int code, bool send_close, std::wstring const& func
 {
 	logger_.log(logmsg::debug_debug, L"tls_layer_impl::failure(%d)", code);
 	if (code) {
-		log_error(code, function);
-		if (socket_eof_) {
-			if (code == GNUTLS_E_UNEXPECTED_PACKET_LENGTH
-#ifdef GNUTLS_E_PREMATURE_TERMINATION
-				|| code == GNUTLS_E_PREMATURE_TERMINATION
-#endif
-				)
-			{
-				if (state_ != socket_state::shut_down || !shutdown_silence_read_errors_) {
-					logger_.log(logmsg::status, server_ ? fztranslate("Client did not properly shut down TLS connection") : fztranslate("Server did not properly shut down TLS connection"));
-				}
+		bool suppress{};
+		auto level = logmsg::error;
+		if (socket_eof_ && code == (GNUTLS_E_UNEXPECTED_PACKET_LENGTH || code == GNUTLS_E_PREMATURE_TERMINATION)) {
+			suppress = state_ == socket_state::shut_down && shutdown_silence_read_errors_;
+			if (!suppress && state_ == socket_state::connected && unexpected_eof_cb_) {
+				suppress = !unexpected_eof_cb_();
 			}
+			if (suppress) {
+				level = logmsg::debug_warning;
+			}
+		}
+		log_error(code, function, level);
+		if (!suppress && socket_eof_ && (code == GNUTLS_E_UNEXPECTED_PACKET_LENGTH || code == GNUTLS_E_PREMATURE_TERMINATION)) {
+			logger_.log(logmsg::status, server_ ? fztranslate("Client did not properly shut down TLS connection") : fztranslate("Server did not properly shut down TLS connection"));
 		}
 	}
 
@@ -1070,7 +1233,7 @@ int tls_layer_impl::shutdown()
 
 	state_ = socket_state::shutting_down;
 
-	if (!send_buffer_.empty()) {
+	if (!send_buffer_.empty() || send_new_ticket_) {
 		logger_.log(logmsg::debug_verbose, L"Postponing shutdown, send_buffer_ not empty");
 		return EAGAIN;
 	}
@@ -1120,8 +1283,10 @@ int tls_layer_impl::continue_shutdown()
 
 void tls_layer_impl::set_verification_result(bool trusted)
 {
+	logger_.log(logmsg::debug_verbose, L"set_verification_result(%s)", trusted ? "true"sv : "false"sv);
+
 	if (state_ != socket_state::connecting && !handshake_successful_) {
-		logger_.log(logmsg::debug_warning, L"TrustCurrentCert called at wrong time.");
+		logger_.log(logmsg::debug_warning, L"set_verification_result called at wrong time.");
 		return;
 	}
 
@@ -1131,19 +1296,19 @@ void tls_layer_impl::set_verification_result(bool trusted)
 	if (trusted) {
 		state_ = socket_state::connected;
 
+#if DEBUG_SOCKETEVENTS
+		if (can_read_from_socket_) {
+			assert(!debug_can_read_);
+			debug_can_read_ = true;
+		}
+		assert(!debug_can_write_);
+		debug_can_write_ = true;
+#endif
 		if (tls_layer_.event_handler_) {
 			tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::connection, 0);
 			if (can_read_from_socket_) {
-#if DEBUG_SOCKETEVENTS
-				assert(!debug_can_read_);
-				debug_can_read_ = true;
-#endif
 				tls_layer_.event_handler_->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
 			}
-#if DEBUG_SOCKETEVENTS
-			assert(!debug_can_write_);
-			debug_can_write_ = true;
-#endif
 		}
 
 		return;
@@ -1169,13 +1334,15 @@ static std::string bin2hex(unsigned char const* in, size_t size)
 }
 
 
-bool tls_layer_impl::extract_cert(gnutls_x509_crt_t const& cert, x509_certificate& out, bool last)
+bool tls_layer_impl::extract_cert(gnutls_x509_crt_t const& cert, x509_certificate& out, bool last, logger_interface * logger)
 {
 	datetime expiration_time(gnutls_x509_crt_get_expiration_time(cert), datetime::seconds);
 	datetime activation_time(gnutls_x509_crt_get_activation_time(cert), datetime::seconds);
 
 	if (!activation_time || !expiration_time || expiration_time < activation_time) {
-		logger_.log(logmsg::error, fztranslate("Could not extract validity period of certificate"));
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Could not extract validity period of certificate"));
+		}
 		return false;
 	}
 
@@ -1211,28 +1378,38 @@ bool tls_layer_impl::extract_cert(gnutls_x509_crt_t const& cert, x509_certificat
 	std::string subject, issuer;
 
 	datum_holder raw_subject;
-	if (!gnutls_x509_crt_get_dn3(cert, &raw_subject, 0)) {
+	res = gnutls_x509_crt_get_dn3(cert, &raw_subject, 0);
+	if (!res) {
 		subject = raw_subject.to_string_view();
 	}
 	else {
-		log_error(res, L"gnutls_x509_crt_get_dn3");
+		if (logger) {
+			logger->log(logmsg::debug_warning, "gnutls_x509_crt_get_dn3 failed with %d", res);
+		}
 	}
 	if (subject.empty()) {
-		logger_.log(logmsg::error, fztranslate("Could not get distinguished name of certificate subject, gnutls_x509_get_dn failed"));
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Could not get distinguished name of certificate subject, gnutls_x509_get_dn failed"));
+		}
 		return false;
 	}
 
 	std::vector<x509_certificate::subject_name> alt_subject_names = get_cert_subject_alt_names(cert);
 
 	datum_holder raw_issuer;
-	if (!gnutls_x509_crt_get_issuer_dn3(cert, &raw_issuer, 0)) {
+	res = gnutls_x509_crt_get_issuer_dn3(cert, &raw_issuer, 0);
+	if (!res) {
 		issuer = raw_issuer.to_string_view();
 	}
 	else {
-		log_error(res, L"gnutls_x509_crt_get_issuer_dn3");
+		if (logger) {
+			logger->log(logmsg::debug_warning, "gnutls_x509_crt_get_issuer_dn3 failed with %d", res);
+		}
 	}
 	if (issuer.empty() ) {
-		logger_.log(logmsg::error, fztranslate("Could not get distinguished name of certificate issuer, gnutls_x509_get_issuer_dn failed"));
+		if (logger) {
+			logger->log(logmsg::error, fztranslate("Could not get distinguished name of certificate issuer, gnutls_x509_get_issuer_dn failed"));
+		}
 		return false;
 	}
 
@@ -1253,7 +1430,9 @@ bool tls_layer_impl::extract_cert(gnutls_x509_crt_t const& cert, x509_certificat
 
 	datum_holder der;
 	if (gnutls_x509_crt_export2(cert, GNUTLS_X509_FMT_DER, &der) != GNUTLS_E_SUCCESS || !der.data || !der.size) {
-		logger_.log(logmsg::error, L"gnutls_x509_crt_get_issuer_dn");
+		if (logger) {
+			logger->log(logmsg::error, L"gnutls_x509_crt_get_issuer_dn");
+		}
 		return false;
 	}
 	std::vector<uint8_t> data(der.data, der.data + der.size);
@@ -1389,6 +1568,29 @@ int tls_layer_impl::get_algorithm_warnings() const
 	return algorithmWarnings;
 }
 
+int tls_layer_impl::load_certificates(std::string_view const& in, bool pem, gnutls_x509_crt_t *& certs, unsigned int & certs_size, bool & sort)
+{
+	gnutls_datum_t dpem;
+	dpem.data = reinterpret_cast<unsigned char*>(const_cast<char *>(in.data()));
+	dpem.size = in.size();
+	unsigned int flags{};
+	if (sort) {
+		flags |= GNUTLS_X509_CRT_LIST_FAIL_IF_UNSORTED;
+	}
+
+	int res = gnutls_x509_crt_list_import2(&certs, &certs_size, &dpem, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, flags);
+	if (res == GNUTLS_E_CERTIFICATE_LIST_UNSORTED) {
+		sort = false;
+		flags |= GNUTLS_X509_CRT_LIST_SORT;
+		res = gnutls_x509_crt_list_import2(&certs, &certs_size, &dpem, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, flags);
+	}
+
+	if (res != GNUTLS_E_SUCCESS) {
+		certs = nullptr;
+		certs_size = 0;
+	}
+	return res;
+}
 
 bool tls_layer_impl::get_sorted_peer_certificates(gnutls_x509_crt_t *& certs, unsigned int & certs_size)
 {
@@ -1404,52 +1606,35 @@ bool tls_layer_impl::get_sorted_peer_certificates(gnutls_x509_crt_t *& certs, un
 	}
 
 	// Convert them all to PEM
-	gnutls_datum_t *pem_cert_list = new gnutls_datum_t[cert_list_size];
+	// Avoid gnutls_pem_base64_encode2, excessive allocations.
+	auto constexpr header = "-----BEGIN CERTIFICATE-----\r\n"sv;
+	auto constexpr footer = "\r\n-----END CERTIFICATE-----\r\n"sv;
+
+	size_t cap = cert_list_size * header.size() + footer.size();
 	for (unsigned i = 0; i < cert_list_size; ++i) {
-		if (gnutls_pem_base64_encode2("CERTIFICATE", cert_list + i, pem_cert_list + i) != 0) {
-			for (unsigned int j = 0; j < i; ++j) {
-				gnutls_free(pem_cert_list[j].data);
-			}
-			delete [] pem_cert_list;
-			logger_.log(logmsg::error, fztranslate("gnutls_pem_base64_encode2 failed"));
-			return false;
-		}
+		cap += ((cert_list[i].size + 2) / 3) * 4;
 	}
 
-	// Concatenate them
-	gnutls_datum_t concated_certs{};
-	for (unsigned i = 0; i < cert_list_size; ++i) {
-		concated_certs.size += pem_cert_list[i].size;
-	}
-	concated_certs.data = new unsigned char[concated_certs.size];
-	concated_certs.size = 0;
-	for (unsigned i = 0; i < cert_list_size; ++i) {
-		memcpy(concated_certs.data + concated_certs.size, pem_cert_list[i].data, pem_cert_list[i].size);
-		concated_certs.size += pem_cert_list[i].size;
-	}
+	std::string pem;
+	pem.reserve(cap);
 
 	for (unsigned i = 0; i < cert_list_size; ++i) {
-		gnutls_free(pem_cert_list[i].data);
+		pem += header;
+		base64_encode_append(pem, to_view(cert_list[i]), base64_type::standard, true);
+		pem += footer;
 	}
-	delete[] pem_cert_list;
 
 	// And now import the certificates
-	int res = gnutls_x509_crt_list_import2(&certs, &certs_size, &concated_certs, GNUTLS_X509_FMT_PEM, GNUTLS_X509_CRT_LIST_FAIL_IF_UNSORTED);
+	bool sort = true;
+	int res = load_certificates(pem, true, certs, certs_size, sort);
 	if (res == GNUTLS_E_CERTIFICATE_LIST_UNSORTED) {
-		logger_.log(logmsg::error, fztranslate("Server sent unsorted certificate chain in violation of the TLS specifications"));
-		res = gnutls_x509_crt_list_import2(&certs, &certs_size, &concated_certs, GNUTLS_X509_FMT_PEM, GNUTLS_X509_CRT_LIST_SORT);
-	}
-
-	delete[] concated_certs.data;
-
-	if (res != GNUTLS_E_SUCCESS) {
-		certs = nullptr;
-		certs_size = 0;
 		logger_.log(logmsg::error, fztranslate("Could not sort peer certificates"));
-		return false;
+	}
+	else if (!sort) {
+		logger_.log(logmsg::error, fztranslate("Server sent unsorted certificate chain in violation of the TLS specifications"));
 	}
 
-	return true;
+	return res == GNUTLS_E_SUCCESS;
 }
 
 void tls_layer_impl::log_verification_error(int status)
@@ -1497,7 +1682,6 @@ void tls_layer_impl::log_verification_error(int status)
 			logger_.log(logmsg::error, fztranslate("Received certificate chain could not be verified. Verification status is %d."), status);
 		}
 	}
-
 }
 
 int tls_layer_impl::verify_certificate()
@@ -1568,6 +1752,7 @@ int tls_layer_impl::verify_certificate()
 	//
 	// In any case, actual trust decision is done later by the user.
 
+	std::vector<x509_certificate> system_trust_chain;
 
 	// First, check system trust
 	if (uses_hostname && system_trust_store_) {
@@ -1575,12 +1760,45 @@ int tls_layer_impl::verify_certificate()
 		auto lease = system_trust_store_->impl_->lease();
 		auto cred = std::get<0>(lease);
 		if (cred) {
+			bool trust_path_ok{true};
+			std::vector<x509_certificate> trust_path;
+			tls_layerCallbacks::verify_output_cb_ = [this, &trust_path_ok, &trust_path](gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output) {
+				if (!trust_path_ok) {
+					return;
+				}
+				if (cert && !issuer && crl && !verification_output) {
+					// Verified against a CRL that the cert isn't expired
+					return;
+				}
+				if (verification_output != 0) {
+					trust_path.clear();
+					return;
+				}
+				if (!issuer || !cert) {
+					trust_path_ok = false;
+					return;
+				}
+
+				x509_certificate info;
+				if (!extract_cert(issuer, info, true, &logger_)) {
+					trust_path_ok = false;
+					return;
+				}
+				if (trust_path.empty() || info.get_fingerprint_sha256() != trust_path.back().get_fingerprint_sha256()) {
+					trust_path.emplace_back(std::move(info));
+				}
+			};
+
+			gnutls_session_set_verify_output_function(session_, c_verify_output_cb);
 			gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cred);
 			unsigned int status = 0;
 			int verifyResult = gnutls_certificate_verify_peers3(session_, to_utf8(hostname_).c_str(), &status);
 			gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cert_credentials_);
 			std::get<1>(lease).unlock();
 
+			gnutls_session_set_verify_output_function(session_, nullptr);
+			tls_layerCallbacks::verify_output_cb_ = nullptr;
+
 			if (verifyResult < 0) {
 				logger_.log(logmsg::debug_warning, L"gnutls_certificate_verify_peers2 returned %d with status %u", verifyResult, status);
 				logger_.log(logmsg::error, fztranslate("Failed to verify peer certificate"));
@@ -1589,8 +1807,26 @@ int tls_layer_impl::verify_certificate()
 			}
 
 			if (!status) {
+				if (!trust_path_ok || trust_path.empty()) {
+					logger_.log(logmsg::error, fztranslate("Failed to extract certificate trust path"));
+					failure(0, true);
+					return EINVAL;
+				}
+
+				// Reverse chain so that it starts at server certificate and add the server cert
+				system_trust_chain.reserve(trust_path.size() + 1);
+				x509_certificate cert;
+				if (!extract_cert(certs.certs[0], cert, false, &logger_)) {
+					failure(0, true);
+					return ECONNABORTED;
+				}
+				system_trust_chain.emplace_back(std::move(cert));
+				for (auto it = trust_path.rbegin(); it != trust_path.rend(); ++it) {
+					system_trust_chain.emplace_back(std::move(*it));
+				}
 				systemTrust = true;
 			}
+			logger_.log(logmsg::debug_verbose, L"System trust store decision: %s", systemTrust ? "true"sv : "false"sv);
 		}
 		else {
 			std::get<1>(lease).unlock();
@@ -1655,14 +1891,12 @@ int tls_layer_impl::verify_certificate()
 			}
 		}
 
-		logger_.log(logmsg::status, fztranslate("Verifying certificate..."));
-
 		std::vector<x509_certificate> certificates;
 		certificates.reserve(certs.certs_size);
 		for (unsigned int i = 0; i < certs.certs_size; ++i) {
 			x509_certificate cert;
-			if (extract_cert(certs.certs[i], cert, i + 1 == certs.certs_size)) {
-				certificates.push_back(cert);
+			if (extract_cert(certs.certs[i], cert, i + 1 == certs.certs_size, &logger_)) {
+				certificates.emplace_back(std::move(cert));
 			}
 			else {
 				failure(0, true);
@@ -1690,7 +1924,7 @@ int tls_layer_impl::verify_certificate()
 					}
 
 					x509_certificate out;
-					if (!extract_cert(issuer, out, true)) {
+					if (!extract_cert(issuer, out, true, &logger_)) {
 						failure(0, true);
 						return ECONNABORTED;
 					}
@@ -1719,10 +1953,11 @@ int tls_layer_impl::verify_certificate()
 			get_mac(),
 			algorithmWarnings,
 			std::move(certificates),
-			systemTrust,
+			std::move(system_trust_chain),
 			hostnameMismatch
 		);
 
+		logger_.log(logmsg::debug_verbose, L"Sending certificate_verification_event");
 		verification_handler_->send_event<certificate_verification_event>(&tls_layer_, std::move(session_info));
 
 		return EAGAIN;
@@ -1756,13 +1991,11 @@ std::string tls_layer_impl::get_key_exchange() const
 	if (dh || ecdh) {
 		char const* const signature_name = gnutls_sign_get_name(static_cast<gnutls_sign_algorithm_t>(gnutls_sign_algorithm_get(session_)));
 		ret = (ecdh ? "ECDHE" : "DHE");
-#if GNUTLS_VERSION_NUMBER >= 0x030600
 		s = gnutls_group_get_name(gnutls_group_get(session_));
 		if (s && *s) {
 			ret += "-";
 			ret += s;
 		}
-#endif
 		if (signature_name && *signature_name) {
 			ret += "-";
 			ret += signature_name;
@@ -1936,11 +2169,24 @@ int tls_layer_impl::connect(native_string const& host, unsigned int port, addres
 	return tls_layer_.next_layer_.connect(host, port, family);
 }
 
+namespace {
+void append_with_size(uint8_t * &p, std::vector<uint8_t> const& d)
+{
+	size_t s = d.size();
+	memcpy(p, &s, sizeof(s));
+	p += sizeof(s);
+	if (s) {
+		memcpy(p, d.data(), s);
+		p += s;
+	}
+}
+}
+
 std::vector<uint8_t> tls_layer_impl::get_session_parameters() const
 {
 	std::vector<uint8_t> ret;
 
-	if (is_client()) {
+	if (!server_) {
 		datum_holder d;
 		int res = gnutls_session_get_data2(session_, &d);
 		if (res) {
@@ -1951,7 +2197,11 @@ std::vector<uint8_t> tls_layer_impl::get_session_parameters() const
 		}
 	}
 	else {
-		ret = ticket_key_;
+		ret.resize(sizeof(size_t) * 3 + ticket_key_.size() + session_db_key_.size() + session_db_data_.size());
+		auto* p = ret.data();
+		append_with_size(p, ticket_key_);
+		append_with_size(p, session_db_key_);
+		append_with_size(p, session_db_data_);
 	}
 
 	return ret;
@@ -2103,6 +2353,116 @@ std::pair<std::string, std::string> tls_layer_impl::generate_selfsigned_certific
 	return ret;
 }
 
+std::pair<std::string, std::string> tls_layer_impl::generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem)
+{
+	std::pair<std::string, std::string> ret;
+
+	gnutls_x509_privkey_t priv;
+	int res = gnutls_x509_privkey_init(&priv);
+	if (res) {
+		return ret;
+	}
+
+	auto fmt = GNUTLS_PK_ECDSA;
+	unsigned int bits = gnutls_sec_param_to_pk_bits(fmt, GNUTLS_SEC_PARAM_HIGH);
+	if (fmt == GNUTLS_PK_RSA && bits < 2048) {
+		bits = 2048;
+	}
+
+	res = gnutls_x509_privkey_generate(priv, fmt, bits, 0);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		return ret;
+	}
+
+	datum_holder kh;
+
+	if (password.empty()) {
+		res = gnutls_x509_privkey_export2(priv, GNUTLS_X509_FMT_PEM, &kh);
+	}
+	else {
+		res = gnutls_x509_privkey_export2_pkcs8(priv, GNUTLS_X509_FMT_PEM, to_utf8(password).c_str(), 0, &kh);
+	}
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		return ret;
+	}
+
+	gnutls_x509_crq_t crq;
+	res = gnutls_x509_crq_init(&crq);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		return ret;
+	}
+
+	res = gnutls_x509_crq_set_version(crq, 3);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	res = gnutls_x509_crq_set_key(crq, priv);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	char const* out{};
+	res = gnutls_x509_crq_set_dn(crq, distinguished_name.c_str(), &out);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	for (auto const& hostname : hostnames) {
+		res = gnutls_x509_crq_set_subject_alt_name(crq, GNUTLS_SAN_DNSNAME, hostname.c_str(), hostname.size(), GNUTLS_FSAN_APPEND);
+		if (res) {
+			gnutls_x509_privkey_deinit(priv);
+			gnutls_x509_crq_deinit(crq);
+			return ret;
+		}
+	}
+
+	res = gnutls_x509_crq_set_key_usage(crq, GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	res = gnutls_x509_crq_set_basic_constraints(crq, 0, -1);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	res = gnutls_x509_crq_sign2(crq, priv, GNUTLS_DIG_SHA256, 0);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	datum_holder ch;
+	res = gnutls_x509_crq_export2(crq, csr_as_pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, &ch);
+	if (res) {
+		gnutls_x509_privkey_deinit(priv);
+		gnutls_x509_crq_deinit(crq);
+		return ret;
+	}
+
+	gnutls_x509_privkey_deinit(priv);
+	gnutls_x509_crq_deinit(crq);
+	ret.first = kh.to_string();
+	ret.second = ch.to_string();
+
+	return ret;
+}
+
 int tls_layer_impl::shutdown_read()
 {
 	if (!can_read_from_socket_) {
@@ -2120,6 +2480,13 @@ int tls_layer_impl::shutdown_read()
 		return ECONNABORTED;
 	}
 
+	if (error == EAGAIN) {
+		can_read_from_socket_ = false;
+#if DEBUG_SOCKETEVENTS
+		debug_can_read_ = false;
+#endif
+	}
+
 	return error;
 }
 
@@ -2131,21 +2498,20 @@ void tls_layer_impl::set_event_handler(event_handler* pEvtHandler, fz::socket_ev
 	tls_layer_.event_handler_ = pEvtHandler;
 
 	if (pEvtHandler) {
-		if ((state_ == socket_state::connected || state_ == socket_state::shutting_down) && !(pending & (socket_event_flag::write | socket_event_flag::connection)) && !(retrigger_block & socket_event_flag::write)) {
+		if (can_write_to_socket_ && (state_ == socket_state::connected || state_ == socket_state::shutting_down) && !(pending & (socket_event_flag::write | socket_event_flag::connection)) && !(retrigger_block & socket_event_flag::write)) {
 			pEvtHandler->send_event<socket_event>(&tls_layer_, socket_event_flag::write, 0);
-		}
-		if ((state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) && !(pending & socket_event_flag::read) && !(retrigger_block & socket_event_flag::read)) {
-			pEvtHandler->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
-		}
-
 #if DEBUG_SOCKETEVENTS
-		if (state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down) {
-			debug_can_read_ = true;
-		}
-		if (state_ == socket_state::connected || state_ == socket_state::shutting_down) {
-			debug_can_write_ = true;
-		}
+			assert(debug_can_write_);
 #endif
+		}
+		if (can_read_from_socket_ && (state_ == socket_state::connected || state_ == socket_state::shutting_down || state_ == socket_state::shut_down)) {
+			if (!(pending & socket_event_flag::read) && !(retrigger_block & socket_event_flag::read)) {
+				pEvtHandler->send_event<socket_event>(&tls_layer_, socket_event_flag::read, 0);
+#if DEBUG_SOCKETEVENTS
+				assert(debug_can_read_);
+#endif
+			}
+		}
 	}
 
 }
@@ -2161,7 +2527,12 @@ bool tls_layer_impl::do_set_alpn()
 		data[i].data = reinterpret_cast<unsigned char *>(const_cast<char*>(alpn_[i].c_str()));
 		data[i].size = alpn_[i].size();
 	}
-	int res = gnutls_alpn_set_protocols(session_, data, alpn_.size(), GNUTLS_ALPN_MANDATORY);
+	int flags = GNUTLS_ALPN_MANDATORY;
+	if (alpn_server_priority_ && server_) {
+		flags |= GNUTLS_ALPN_SERVER_PRECEDENCE;
+	}
+	int res = gnutls_alpn_set_protocols(session_, data, alpn_.size(), flags);
+	delete [] data;
 
 	if (res) {
 		log_error(res, L"gnutls_alpn_set_protocols");
@@ -2180,4 +2551,63 @@ std::string tls_layer_impl::get_alpn() const
 	return {};
 }
 
+void tls_layer_impl::set_min_tls_ver(tls_ver ver)
+{
+	min_tls_ver_ = ver;
+}
+
+void tls_layer_impl::set_max_tls_ver(tls_ver ver)
+{
+	max_tls_ver_ = ver;
+}
+
+int tls_layer_impl::new_session_ticket()
+{
+	if (state_ == socket_state::shutting_down || state_ == socket_state::shut_down) {
+		return ESHUTDOWN;
+	}
+	else if (state_ != socket_state::connected) {
+		return ENOTCONN;
+	}
+
+	if (!server_) {
+		return EINVAL;
+	}
+
+	if (gnutls_protocol_get_version(session_) != GNUTLS_TLS1_3) {
+		return 0;
+	}
+
+#if DEBUG_SOCKETEVENTS
+	assert(debug_can_write_);
+	assert(!has_pending_event(tls_layer_.event_handler_, &tls_layer_, socket_event_flag::write));
+#endif
+
+	if (!send_buffer_.empty()|| send_new_ticket_) {
+		send_new_ticket_ = true;
+		return 0;
+	}
+
+	int res = GNUTLS_E_AGAIN;
+	while ((res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN) && can_write_to_socket_) {
+		res = gnutls_session_ticket_send(session_, 1, 0);
+	}
+
+	if (res == GNUTLS_E_SUCCESS) {
+		return 0;
+	}
+	else if (res == GNUTLS_E_AGAIN) {
+		send_new_ticket_ = true;
+		return 0;
+	}
+
+	failure(res, false, L"gnutls_session_ticket_send");
+	return socket_error_ ? socket_error_ : ECONNABORTED;
+}
+
+void tls_layer_impl::set_unexpected_eof_cb(std::function<bool()> && cb)
+{
+	unexpected_eof_cb_ = std::move(cb);
+}
+
 }
