@@ -1,19 +1,19 @@
 #ifndef LIBFILEZILLA_TEST_UTILS_HEADER
 #define LIBFILEZILLA_TEST_UTILS_HEADER
 
-#include "libfilezilla/string.hpp"
+#include "../lib/libfilezilla/string.hpp"
 
 #include <cppunit/TestAssert.h>
 #include <cppunit/extensions/HelperMacros.h>
 
 template<typename T>
-std::string inline value_to_string(T const& t, typename std::enable_if_t<std::is_enum<T>::value>* = nullptr)
+std::string inline value_to_string(T const& t, typename std::enable_if_t<std::is_enum_v<T>>* = nullptr)
 {
 	return std::to_string(std::underlying_type_t<T>(t));
 }
 
 template<typename T>
-std::string inline value_to_string(T const& t, typename std::enable_if_t<std::is_arithmetic<T>::value>* = nullptr)
+std::string inline value_to_string(T const& t, typename std::enable_if_t<std::is_arithmetic_v<T>>* = nullptr)
 {
 	return std::to_string(t);
 }
